"""timesheet hooks: compute progress, completing what's already done in
workorder.
"""

__docformat__ = "restructuredtext en"

from cubicweb.predicates import is_instance
from cubicweb.server import hook

from cubicweb_workorder.hooks import UpdateProgressOp


class UpdateProgressOnActivityModification(hook.Hook):
    __regid__ = 'timesheet.progress.activity'
    __select__ = hook.Hook.__select__ & is_instance('Activity')
    events = ('after_update_entity', )

    def __call__(self):
        if 'duration' in self.entity.cw_edited:
            for workorder in self.entity.done_for:
                UpdateProgressOp.get_instance(self._cw).add_data(
                    (workorder.eid, True))


class UpdateProgressOnDoneForChange(hook.Hook):
    __regid__ = 'timesheet_update_done_for'
    __select__ = (hook.Hook.__select__
                  & hook.match_rtype('done_for', toetypes=('WorkOrder',)))
    events = ('after_add_relation', 'after_delete_relation',)

    def __call__(self):
        UpdateProgressOp.get_instance(self._cw).add_data(
            (self.eidto, True))



# update default activity workflow
wf = rql('Workflow X WHERE A name "Activity", A default_workflow X').one()

# create state archived and transition archive
validated = wf.state_by_name('validated')
archived = wf.add_state(_('archived'))
wf.add_transition(_('archive'), validated, archived, ('managers',))

# allow owners of workorders to validate activities
validate = wf.transition_by_name('validate')
cond = create_entity('RQLExpression', exprtype=u'ERQLExpression',
                     mainvars=u'X', expression=u'X done_for W, W owned_by U',
                     reverse_condition=validate)

# done
commit()



# -*- coding: utf-8 -*-
"""activity related views.

:organization: Logilab
:copyright: 2007-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"


from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb.utils import UStringIO
from cubicweb.schema import display_name
from cubicweb.predicates import (is_instance, multi_columns_rset,
                                 match_user_groups, score_entity, rql_condition)
from cubicweb_web.view import AnyRsetView, EntityView, EntityAdapter
from cubicweb_web import action
from cubicweb_web.component import Link
from cubicweb_web.views import tableview, calendar, navigation, editcontroller

from cubicweb_calendar.views import get_date_range_from_reqform


class ActivitySummaryView(AnyRsetView):
    __regid__ = 'actsummary'
    __select__ = is_instance('Activity') & multi_columns_rset(7)

    # XXX we should make a very strict selector here
    def call(self):
        total_duration = sum(e.duration for e in self.cw_rset.entities())
        self.w(u'<h3>%s: %s</h3>' % (_('total'), total_duration))
        self.wview('table', self.cw_rset, 'null', displaycols=range(1, 6))

        resdict = {}
        for __, __, res, __, duration, __, login in self.cw_rset:
            resdur = resdict.get(login, 0)
            resdur += duration
            resdict[login] = resdur
        self.w(u'<h2>%s</h2>' % _('statistics'))
        self.w(u'<table class="listing">')
        self.w(u'<tr><th>%s</th><th>%s</th></tr>' % (_('resource'), _('duration')))
        for even_odd, (login, resdur) in enumerate(sorted(resdict.iteritems())):
            self.w(u'<tr class="%s">' % (even_odd % 2 and "odd" or "even"))
            self.w(u'<td>%s</td>' % login)
            self.w(u'<td>%s</td>' % resdur)
            self.w(u'</tr>')
        self.w(u'</table>')


class ActivitySubmitView(EntityView):
    __regid__ = 'activities-list'
    __select__ = is_instance('Resource')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        firstday, lastday = get_date_range_from_reqform(self._cw.form)
        rql = 'Any A WHERE A is Activity, A done_by R, R eid %(r)s'
        if lastday is None:
            rql += ', A diem %(fd)s'
        else:
            rql += ', A diem >= %(fd)s, A diem <= %(ld)s'
        rset = self._cw.execute(rql, {'r': entity.eid,
                                      'fd': firstday, 'ld': lastday})
        self.wview('activitytable', rset, 'null')
        self.wview('activities-submitform',
                   rset=self.cw_rset, row=row, col=col,
                   initargs={'date': firstday})


# XXX see generic definition for tablecontext view in gingouz.views
class ActivityTableContext(EntityView):
    __regid__ = 'tablecontext'
    __select__ = is_instance('Activity')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<a href="%s"><img alt="%s" src="data/accessories-text-editor.png" /></a>' %
               (xml_escape(entity.absolute_url(vid='edition')),
                self._cw._('actions_edit')))


class ActivityTableView(EntityView):
    __regid__ = 'activitytable'
    __select__ = is_instance('Activity')
    title = _('activitytable')

    def call(self, showresource=True):
        eids = ','.join(str(row[0]) for row in self.cw_rset)
        rql = ('Any R, D, DUR, WO, DESCR, S, A, SN, RT, WT ORDERBY D DESC '
               'WHERE '
               '   A is Activity, A done_by R, R title RT, '
               '   A diem D, A duration DUR, '
               '   A done_for WO, WO title WT, '
               '   A description DESCR, A in_state S, S name SN, A eid IN (%s)' % eids)
        rset = self._cw.execute(rql)
        self.wview('activity-table', rset, 'null',
                   showresource=showresource)


class GenericActivityTable(tableview.RsetTableView):
    __regid__ = 'generic-activitytable'
    __select__ = multi_columns_rset()
    title = _('activitytable')
    cellvids = {4: 'editable-final'}
    finalvid = 'editable-final'
    layout_args = {
        'display_filter': 'top',
        'add_view_actions': False,
    }

    def call(self, title=None):
        strio = UStringIO()
        self.paginate(self, w=strio.write, page_size=20)
        super(GenericActivityTable, self).call()
        self.w(strio.getvalue())


class ActivityTable(tableview.RsetTableView):
    __regid__ = 'activity-table'

    def call(self, title=None, showresource=True):
        _ = self._cw._
        self.headers = [_("diem"), _("duration"), _("workpackage"),
                        _("description"), _("state"), ""]
        if showresource:
            self.displaycols = range(7)
            self.headers.insert(0, display_name(self._cw, 'Resource'))
            self.cellvids = {1: 'editable-final',
                             2: 'editable-final',
                             3: 'editable-final',
                             4: 'editable-final'}
        else:  # skip resource column if asked to
            self.displaycols = range(1, 7)
            self.cellvids = {
                             0: 'editable-final',
                             1: 'editable-final',
                             2: 'editable-final',
                             3: 'editable-final'}
        super(ActivityTable, self).call()


class ActivityCalendarItemView(calendar.CalendarItemView):
    __select__ = is_instance('Activity')

    def cell_call(self, row, col):
        activity = self.cw_rset.get_entity(row, col)
        self.w(u'<a href="%s">%s %s</a>' % (xml_escape(activity.absolute_url()),
                                            xml_escape(activity.workorder.dc_long_title()),
                                            activity.duration))


class ValidateActivitiesAction(action.Action):
    __regid__ = 'validate-activities'
    __select__ = (action.Action.__select__
                  & is_instance('Activity')
                  & (match_user_groups('managers')
                     | rql_condition('X done_for W, W owned_by U'))
                  & score_entity(lambda x: x.cw_adapt_to('IWorkflowable').state == 'pending'))

    category = 'mainactions'
    title = _('validate activities')
    trname = 'validate'

    def fill_menu(self, box, menu):
        if self.cw_row is None:
            eids = [row[0] for row in self.cw_rset]
        else:
            eids = (self.cw_rset[self.cw_row][self.cw_col or 0],)
        menu.append(self.build_link(self.title, eids))

    def build_link(self, title, eids, item=None):
        rql = ('INSERT TrInfo X: '
               'X by_transition BT, X wf_info_for Y '
               'WHERE Y eid in ({0}), Y in_state S, S state_of W, '
               'BT transition_of W, BT name \'{1}\'')
        rql = rql.format(','.join(str(eid) for eid in eids), self.trname)
        self._cw.add_js('cubes.timesheet.js')
        href = self._cw.build_url('rqlio/1.0')
        attrs = {'class': 'rqlio',
                 'data-rql': rql}
        return Link(href, self._cw._(title), **attrs)


class ArchiveActivitiesAction(ValidateActivitiesAction):
    __regid__ = 'archive-activities'
    __select__ = (action.Action.__select__
                  & match_user_groups('managers')
                  & is_instance('Activity')
                  & score_entity(lambda x: x.cw_adapt_to('IWorkflowable').state == 'validated'))

    title = _('archive activities')
    trname = 'archive'


WORKORDER_CLOSED_STATES = ['validated', 'canceled']


class MoveActivitiesAction(action.Action):
    __regid__ = 'move-activities'
    __select__ = (action.Action.__select__
                  & ~match_user_groups('managers')
                  & is_instance('Activity')
                  & rql_condition('X done_for W, W owned_by U')
                  & score_entity(lambda x: x.cw_adapt_to('IWorkflowable').state == 'pending'))

    category = 'mainactions'
    submenu = _('move activities')

    def fill_menu(self, box, menu):
        self._cw.add_js('cubes.timesheet.js')
        if self.cw_row is None:
            eids = [str(row[self.cw_col or 0]) for row in self.cw_rset]
        else:
            eids = [str(self.cw_rset[self.cw_row][self.cw_col or 0])]
        for item in self.actual_actions():
            menu.append(self.build_link(item.dc_long_title(), item, eids))

    def actual_actions(self):
        states = ",".join('"%s"' % state for state in WORKORDER_CLOSED_STATES)
        return self._cw.execute(
            "Any W,WT,O,OT ORDERBY OT,WT WHERE W is WorkOrder, "
            f"W in_state S, NOT S name IN ({states}), "
            "O split_into W, O title OT, W title WT, "
            "W owned_by U, U eid %(eid)s",
            {"eid": self._cw.cnx.user.eid},
        ).entities()

    def build_link(self, title, workorder, eids):
        rql = 'SET X done_for Y WHERE X eid IN({0}), Y eid {1}'
        rql = rql.format(','.join(eids), workorder.eid)
        href = self._cw.build_url('rqlio/1.0')
        attrs = {'class': 'rqlio',
                 'data-rql': rql, }
        return Link(href, title, **attrs)


class ManagerMoveActivitiesAction(MoveActivitiesAction):
    __select__ = (action.Action.__select__
                  & match_user_groups('managers')
                  & is_instance('Activity')
                  & score_entity(lambda x: x.cw_adapt_to('IWorkflowable').state != 'archived'))

    def actual_actions(self):
        return self._cw.execute('Any W,WT,O,OT ORDERBY OT,WT WHERE W is WorkOrder, '
                                'O split_into W, O title OT, W title WT'
                                ).entities()


class ActivityIPrevNextAdapter(navigation.IPrevNextAdapter):
    __select__ = is_instance('Activity')

    def previous_entity(self):
        entity = self.entity
        execute = self._cw.execute
        # if the smallest duration
        rset = execute("Activity A ORDERBY DUR DESC LIMIT 1 WHERE "
                       "A done_by R, R euser U, U eid %(u)s, A diem %(d)s, "
                       "A duration DUR, A duration < %(t)s",
                       {'u': entity.user.eid, 'd': entity.diem,
                        't': entity.duration})
        if rset:
            return rset.get_entity(0, 0)
        # the smallest id
        rset = execute("Activity A ORDERBY A DESC LIMIT 1 WHERE "
                       "A done_by R, R euser U, U eid %(u)s, A diem %(d)s, "
                       "A duration < %(t)s, A eid < %(eid)s",
                       {'u': entity.user.eid, 'd': entity.diem,
                        't': entity.duration, 'eid': entity.eid})
        if rset:
            return rset.get_entity(0, 0)
        # next days
        rset = execute("Activity A ORDERBY D DESC LIMIT 1 WHERE "
                       "A done_by R, R euser U, U eid %(u)s, "
                       "A diem D, A diem < %(d)s ",
                       {'u': entity.user.eid, 'd': entity.diem})
        if rset:
            return rset.get_entity(0, 0)

    def next_entity(self):
        entity = self.entity
        execute = self._cw.execute
        rset = execute("Activity A ORDERBY DUR LIMIT 1 WHERE "
                       "A done_by R, R euser U, U eid %(u)s, A diem %(d)s, "
                       "A duration DUR, A duration > %(t)s",
                       {'u': entity.user.eid, 'd': entity.diem,
                        't': entity.duration})
        if rset:
            return rset.get_entity(0, 0)
        rset = execute("Activity A ORDERBY A LIMIT 1 WHERE "
                       "A done_by R, R euser U, U eid %(u)s, A diem %(d)s, "
                       "A eid > %(eid)s, A duration > %(t)s",
                       {'u': entity.user.eid, 'd': entity.diem,
                        't': entity.duration, 'eid': entity.eid})
        if rset:
            return rset.get_entity(0, 0)
        rset = execute("Activity A ORDERBY D LIMIT 1 WHERE "
                       "A done_by R, R euser U, U eid %(u)s, "
                       "A diem D, A diem > %(d)s ",
                       {'u': entity.user.eid, 'd': entity.diem})
        if rset:
            return rset.get_entity(0, 0)


class ActivityIEditControlAdapter(editcontroller.IEditControlAdapter):
    __select__ = is_instance('Activity')

    def after_deletion_path(self):
        """return (path, parameters) which should be used as redirect
        information when this entity is being deleted
        """
        return 'view', {}


class ActivityICalendarViewsAdapter(EntityAdapter):
    """calendar views interface"""
    __regid__ = 'ICalendarViews'
    __select__ = is_instance('Activity')

    def matching_dates(self, begin, end):
        """calendar views interface"""
        return [self.entity.diem]



"""gingouz specific boxes

:organization: Logilab
:copyright: 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from datetime import date

from logilab.common.date import first_day
from logilab.common.registry import objectify_predicate
from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb import tags
from cubicweb_web import component


@objectify_predicate
def has_resource(cls, req, **kwargs):
    return bool(req.user.cw_adapt_to('timesheet.IResource'))


class ActivitiesBox(component.CtxComponent):
    """display a box with an activity calendar"""
    __regid__ = 'timesheet.activities_box'
    __select__ = component.CtxComponent.__select__ & has_resource()
    visible = True  # enabled by default
    title = _('My activities')
    order = 10

    def render_title(self, w):
        title = self._cw._(self.title)
        url = self._cw.build_url(rql="Any C WHERE R use_calendar CU, CU use_calendar C, "
                                 "R euser U, U eid %s" % self._cw.user.eid)
        title += u'&nbsp;&nbsp;<a href="%s" title="%s"><img alt="calendar icon" src="%s"/> </a>' % (
            xml_escape(url), self._cw._('see my calendars'),
            xml_escape(self._cw.data_url('office-calendar.png')))
        w(title)

    def render_body(self, w):
        resource = self._cw.user.cw_adapt_to('timesheet.IResource').resource
        resource.view('activity_calendar', firstday=first_day(date.today()),
                      calid='selftid', w=w)


class MyActionsBox(component.CtxComponent):
    """display a box with util links"""
    __regid__ = 'timesheet.my_actions_box'
    __select__ = component.CtxComponent.__select__ & has_resource()

    visible = True  # enabled by default
    title = _('My actions')
    order = 3

    def render_body(self, w):
        req = self._cw
        resource = req.user.cw_adapt_to('timesheet.IResource').resource
        self.append(tags.a(req._('my_space'), href=resource.absolute_url()))
        # select default calendar for currently logged user
        if resource.default_calendar:
            cal = resource.default_calendar
            action = req.vreg['actions'].select('ask-off-days', req,
                                                rset=cal.as_rset())
            self.append(tags.a(req._(action.title), href=action.url()))
        self.render_items(w)



"""gingouz specific primary views
"""

from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb.schema import display_name
from cubicweb_web.view import EntityView
from cubicweb.predicates import is_instance, one_line_rset
from cubicweb_web.views import primary, tabs

from cubicweb_calendar.views import primaries as cal_primaries


def print_resource_ratio(self, rset):
    """prints the ratio time spent per resource / time spent by everyone
    """
    if rset:
        duration = sum(duration for r, s, duration, i, di, mi, ma in rset if duration)
        tot_by_res = dict.fromkeys((row[0] for row in rset), 0)
        for r, s, d, i, di, mi, ma in rset:
            tot_by_res[r] += d
        done = set()
        self.w(u'<p>%s</p>' % self._cw._('percentage of time spent per resource'))
        self.w(u'<table class="listing">')
        self.w(u'<tr><th>%s</th><th>%s</th></tr>' % (display_name(self._cw, 'Resource'),
                                                     self._cw._('time spent ratio')))
        for row in range(len(rset)):
            if rset[row][0] not in done:
                self.w(u'<tr>')
                entity = rset.get_entity(row, 0)
                eid = entity.eid
                done.add(eid)
                ratio = tot_by_res[eid] * 100 / duration
                self.w(u'<td>%s</td><td>%.2f %%</td>' % (entity.view('oneline'), ratio))
                self.w(u'</tr>')
        self.w(u'</table>')


# order #####################################################################
class OrderPrimaryView(tabs.TabsMixin, primary.PrimaryView):
    __select__ = is_instance('Order') & one_line_rset()
    tabs = [_('activity-overview'), _('activity-details')]
    default_tab = 'activity-overview'

    def render_entity_title(self, entity):
        self.w(u'<h1>%s</h1>' % xml_escape(entity.dc_long_title()))
        if entity.date:
            self.w(u'%s: %s<br />' % (display_name(self._cw, 'date'),
                                      self._cw.format_date(entity.date)))

    def render_entity(self, entity):
        self.render_entity_title(entity)
        rql = 'Any WO ORDERBY T WHERE W eid %(eid)s, W split_into WO, WO title T'
        rset = self._cw.execute(rql, {'eid': entity.eid})
        self.wview('ic_progress_table_view', rset, 'null')
        self.render_tabs(self.tabs, self.default_tab, entity)


class OrderOverviewView(EntityView):
    __regid__ = 'activity-overview'
    __select__ = is_instance('Order')
    title = _('activity-overview')

    def cell_call(self, row, col):
        _ = self._cw._
        entity = self.cw_rset.get_entity(row, col)
        rset = self._cw.execute(entity.rql_activities_groupby_resource, {'eid': entity.eid})
        headers = (_('Resource'), _('status'), _('time spent'), _('begin date'), _('end date'))
        self.wview('table', rset, 'null', headers=headers)
        self.w('<div style="margin-top: 1em"><a href="%s">%s</a></div>' %
               (self._cw.build_url(rql='Any A WHERE A done_for Y, X split_into Y, X eid %s' % entity.eid),
                self._cw._('View all activities')))


class OrderDetailView(EntityView):
    __regid__ = 'activity-details'
    title = _('activity-details')
    __select__ = is_instance('Order') & one_line_rset()

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rset = self._cw.execute(entity.rql_activities, {'eid': entity.eid})
        self.wview('generic-activitytable', rset, 'null')

# workorder ####################################################################


class WorkorderPrimaryView(tabs.TabsMixin, primary.PrimaryView):
    """display workpackage's activity summary"""
    __select__ = is_instance('WorkOrder') & one_line_rset()
    tabs = [_('activity-overview'), _('activity-details')]
    default_tab = 'activity-overview'

    def render_entity_title(self, entity):
        self.w(u'<h1>%s</h1>' % xml_escape(entity.dc_long_title()))

    def render_entity(self, entity):
        self.render_entity_title(entity)
        view = self._cw.vreg['views'].select('progress_table_view', self._cw,
                                             rset=entity.as_rset())
        columns = list(view.columns)
        for col in ('project', 'milestone'):
            try:
                columns.remove(col)
            except ValueError:
                self.warning('could not remove %s from columns' % col)
        view.render(w=self.w, columns=columns)
        rset = self._cw.execute(
            'Any R,B,E,C WHERE WO eid %(eid)s, C commit_for WO, '
            'C duration D, C begin_date B, C end_date E, '
            'C commit_by R', {'eid': entity.eid})
        if rset:
            self.w('<h2>Engagements</h2>')
            self.wview('table', rset, 'null')
            rset_total = self._cw.execute(
                'Any SUM(D) GROUPBY WO WHERE '
                'WO eid %(eid)s, C commit_for WO, C duration D',
                {'eid': entity.eid})
            total = rset_total[0][0]
            self.w('<p>Total engagements est %s' % total)
            self.w(' pour budget de %s donc il reste %s</p>'
                   % (entity.budget, entity.budget-total))
        self.render_tabs(self.tabs, self.default_tab, entity)


class WorkOrderOverviewView(EntityView):
    __regid__ = 'activity-overview'
    __select__ = is_instance('WorkOrder')
    title = _('activity-overview')

    def cell_call(self, row, col):
        _ = self._cw._
        entity = self.cw_rset.get_entity(row, col)
        rset = self._cw.execute(entity.rql_activities_groupby_resource, {'eid': entity.eid})
        headers = (_('Resource'), _('status'), _('time spent'), _('begin date'), _('end date'))
        self.wview('table', rset, 'null', headers=headers)
        self.w('<div style="margin-top: 1em"><a href="%s">%s</a></div>' %
               (self._cw.build_url(rql='Any A WHERE A done_for X, X eid %s' % entity.eid),
                self._cw._('view all activities')))


class WorkOrderDetailView(EntityView):
    __regid__ = 'activity-details'
    __select__ = is_instance('WorkOrder') & one_line_rset()
    title = _('activity-details')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rset = self._cw.execute(entity.rql_activities, {'eid': entity.eid})
        self.wview('generic-activitytable', rset, 'null')

# resource #####################################################################


class ResourcePrimaryView(tabs.TabsMixin, primary.PrimaryView):
    """display workpackage's activity summary"""
    __select__ = is_instance('Resource') & one_line_rset()
    tabs = [_('activity-current'), _('activity-overview'), _('activity-details'),
            _('resource-calendar'), ]
    default_tab = 'activity-current'

    def render_entity_title(self, entity):
        self.w(u'<h1>%s</h1>' % xml_escape(entity.dc_long_title()))

    def render_entity(self, entity):
        self.render_entity_title(entity)
        entity.view('activities-submitform', w=self.w)
        self.render_tabs(self.tabs, self.default_tab, entity)


class ResourceRecentView(EntityView):
    __regid__ = 'activity-current'
    title = _('recent activities')
    __select__ = is_instance('Resource')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<h2>%s</h2>' % _('Recent activities'))
        rset = self._cw.execute('Any A ORDERBY D DESC LIMIT 12 '
                                'WHERE A done_by R, R eid %(r)s, A diem D',
                                {'r': entity.eid})
        self.wview('activitytable', rset, 'null', showresource=False)
        self.w(u'<h2>%s</h2>' % _('Open workorders'))
        rset = self._cw.execute('Any W WHERE W is WorkOrder, W todo_by R, '
                                'R eid %(r)s, W in_state S, S name "in progress"',
                                {'r': entity.eid})
        self.wview('ic_progress_table_view', rset, 'null')


class ResourceOverviewView(EntityView):
    __regid__ = 'activity-overview'
    __select__ = is_instance('Resource')
    title = _('activity overview')

    def cell_call(self, row, col):
        _ = self._cw._
        entity = self.cw_rset.get_entity(row, col)
        rset = self._cw.execute(entity.rql_activities_groupby_workorder, {'eid': entity.eid})
        headers = (_('workorder'), _('status'), _('time spent'), _('begin date'), _('end date'))
        self.wview('table', rset, 'null', headers=headers, cellvids={0: 'outofcontext'})  # , subvid='tablecontext')


class ResourceDetailView(EntityView):
    __regid__ = 'activity-details'
    __select__ = is_instance('Resource') & one_line_rset()
    title = _('activity details')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rset = self._cw.execute(entity.rql_activities,  {'eid': entity.eid})
        self.wview('generic-activitytable', rset, 'null')


class ResourceCalendarView(EntityView):
    __regid__ = 'resource-calendar'
    __select__ = is_instance('Resource')
    title = _('resource-calendar')

    def cell_call(self, row, col):
        _ = self._cw._
        entity = self.cw_rset.get_entity(row, col)
        # calendars
        self.w(u'<h2>%s</h2>' % _('Calendars'))
        rset = self._cw.execute('Any A,B,C ORDERBY A WHERE U eid %(eid)s, '
                                'U use_calendar CU, CU use_calendar C, '
                                'CU start A, CU stop B', {'eid': entity.eid})
        headers = (_('begin date'), _('end date'), _('calendar'))
        self.wview('table', rset, 'null', headers=headers)


# activity  ######################################################################

class ActivityPrimaryView(primary.PrimaryView):
    __select__ = is_instance('Activity')

    def cell_call(self, row, col):
        _ = self._cw._
        activity = self.cw_rset.get_entity(row, col)
        self.w(u'<h2>%s: %s</h2>' % (
            display_name(self._cw, 'State'),
            xml_escape(activity.cw_adapt_to('IWorkflowable').printable_state)))
        rql = ("Any D,U,DU,WO,DE,A WHERE A is Activity, A done_by R, R euser U, "
               "A diem D, A done_for WO, A eid %(x)s, A duration DU, A description DE")
        headers = (_("diem"), _("euser"), _("duration"), _("workpackage"), _("description"), u"")
        rset = self._cw.execute(rql, {'x': activity.eid})
        if rset:
            self.wview('table', rset, headers=headers, displaycols=range(0, 6))  # , subvid='tablecontext')
        else:
            # missing information or remote sources unavailable
            super(ActivityPrimaryView, self).cell_call(row, col)

# calendar  ####################################################################


class PrimaryCalendar(cal_primaries.PrimaryCalendar):

    def render_entity_attributes(self, entity):
        super(PrimaryCalendar, self).render_entity_attributes(entity)
        rset = self._cw.execute('Any A,B,U ORDERBY A WHERE C eid %(eid)s, '
                                'U use_calendar CU, CU use_calendar C, '
                                'CU start A, CU stop B', {'eid': entity.eid})
        if rset:
            self.w(u'<h2>%s</h2>' % _('Resources using this calendar'))
            headers = (_('begin date'), _('end date'), _('resource'))
            self.wview('table', rset, 'null', headers=headers)

# registration #################################################################


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__,
                      butclasses=(PrimaryCalendar,))
    vreg.register_and_replace(PrimaryCalendar, cal_primaries.PrimaryCalendar)



"""template-specific forms/views/actions/components"""
from cubicweb_web.views import uicfg
from cubicweb_web.views.urlrewrite import SimpleReqRewriter, rgx

uicfg.primaryview_section.tag_object_of(('*', 'use_calendar', 'Calendar'), 'hidden')

uicfg.actionbox_appearsin_addmenu.tag_subject_of(('Resource', 'has_vacation', '*'), True)
uicfg.actionbox_appearsin_addmenu.tag_object_of(('Resource', 'euser', 'CWUser'), True)
uicfg.actionbox_appearsin_addmenu.tag_object_of(('Commitment', 'commit_for', 'WorkOrder'), True)

uicfg.reledit_ctrl.tag_subject_of(('Activity', 'done_for', '*'),
                                  {'rvid': 'outofcontext'})


rql_all_users_resources = (r'Any R,L ORDERBY L WHERE R is Resource, R euser U,'
                           r' U login L,U in_state S, S name "activated"')

rql_user_or_group_resources = (r'Any R WHERE R is Resource, R euser U,'
                               r' EXISTS (U login "\1")'
                               r' OR EXISTS(U in_group G, G name "\1")')


class TimesheetReqRewriter(SimpleReqRewriter):
    rules = [

        # activities

        (rgx(r'/activities/(.*?)/(\d{4})$'),
         dict(year=r'\2', month=r'01', day=r'01',
              vid='activities-submit',
              rql=(r'Any A ORDERBY D WHERE A is Activity, A diem D,'
                   r' A diem >= "\2-01-01", A diem <= "\2-12-31", A done_by R,'
                   r' R euser U, U login "\1"'))),

        (rgx(r'/activities/(.*?)/(\d{4})-(\d\d)'),
         dict(year=r'\2', month=r'\3', day=r'01', vid='activities-submit',
              rql=(r'Any A WHERE A is Activity, A diem "\2-\3-01",'
                   r' A done_by R, R euser U, U login "\1"'))),

        (rgx(r'/activities/(.*?)/(\d{8})-(\d{8})'),
         dict(vid='activities-list', firstday=r'\2', lastday=r'\3',
              rql=r'Any R WHERE R is Resource, R title "\1"')),

        (rgx(r'/activities/(.*?)/(\d{8})$'),
         dict(vid='activities-list', firstday=r'\2',
              rql=r'Any R WHERE R is Resource, R title "\1"')),

        # this rule is for backward compatibility
        (rgx(r'/activities/(.*?)/(\d{4})-(\d\d)-(\d\d)'),
         dict(year=r'\2', month=r'\3', day=r'\4', vid='activities-submit',
              rql=(r'Any A WHERE A is Activity, A diem "\2-\3-\4", A done_by R,'
                   r' R euser U, U login "\1"'))),

        # missing-activites

        (rgx(r'/missing-activities/all/(\d{8})'),
         dict(vid='missing-activities', firstday=r'\1',
              rql=rql_all_users_resources)),

        (rgx(r'/missing-activities/all/(\d{8})-(\d{8})'),
         dict(vid='missing-activities', firstday=r'\1', lastday=r'\2',
              rql=rql_all_users_resources)),

        (rgx(r'/missing-activities/all/(\d{8})-[tT][oO][dD][aA][yY]'),
         dict(vid='missing-activities', firstday=r'\1', lastday='TODAY',
              rql=rql_all_users_resources)),

        (rgx(r'/missing-activities/(.*?)/(\d{8})'),
         dict(vid='missing-activities', firstday=r'\2',
              rql=rql_user_or_group_resources)),

        (rgx(r'/missing-activities/(.*?)/(\d{8})-(\d{8})'),
         dict(vid='missing-activities', firstday=r'\2', lastday=r'\3',
              rql=rql_user_or_group_resources)),

        (rgx(r'/missing-activities/(.*?)/(\d{8})-[tT][oO][dD][aA][yY]'),
         dict(vid='missing-activities', firstday=r'\2', lastday='TODAY',
              rql=rql_user_or_group_resources)),

        # commitments

        (rgx(r'/commitments/(.*?)$'),
         dict(vid='resource.commitment-dashboard',
              rql=r'Any R WHERE R is Resource, R title "\1"')),


        ]



# -*- coding: utf-8 -*-
"""commitment related views.

:organization: Logilab
:copyright: 2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"


from cubicweb.predicates import is_instance

from cubicweb_web.view import EntityView
from cubicweb_web.views import tableview


class ResourceCommitmentDashboard(EntityView):
    __regid__ = 'resource.commitment-dashboard'
    __select__ = is_instance('Resource')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rql = ('Any R, WO, BD, ED, DU, C ORDERBY WOT '
               'WHERE '
               '   R eid %(r)s, R is Resource, '
               '   C commit_by R, C is Commitment, '
               '   C duration DU, '
               '   C begin_date BD, C end_date ED, '
               '   C commit_for WO, WO title WOT')
        rset = self._cw.execute(rql, {'r': entity.eid})
        self.wview('commitment.table', rset, 'null')
        self.wview('commitment.submitform',
                   rset=self.cw_rset, row=row, col=col,
                   initargs={})


class CommitmentTable(tableview.RsetTableView):
    __regid__ = 'commitment.table'

    def call(self):
        self.cellvids = {
            1: 'outofcontext',
            2: 'editable-final',
            3: 'editable-final',
            4: 'editable-final',
        }
        super(CommitmentTable, self).call()



# ResourceType person
from cubicweb import _

restype = rql('INSERT Resourcetype RT: RT title "person"').rows[0][0]

# Activity workflow
wf = add_workflow('activities default workflow', 'Activity')
pending = wf.add_state(_('pending'), initial=True)
validated = wf.add_state(_('validated'))
archived = wf.add_state(_('archived'))
wf.add_transition(_('validate'), pending, validated, ('managers',), 'X done_for W, W owned_by U')
wf.add_transition(_('archive'), validated, archived, ('managers',))



import six
from cubicweb_web import INTERNAL_FIELD_VALUE, formwidgets as fw
from cubicweb_web.views import uicfg

_afs = uicfg.autoform_section
_affk = uicfg.autoform_field_kwargs

_afs.tag_subject_of(('Resource', 'use_calendar', '*'), 'main', 'inlined')
_afs.tag_subject_of(('WorkOrder', 'uses_technology', '*'), 'main', 'attributes')
_afs.tag_subject_of(('WorkOrder', 'uses_technology', '*'), 'muledit', 'attributes')
# XXX should not be always hidden
_affk.tag_subject_of(('Activity', 'done_by', '*'), {'widget': fw.HiddenInput})


def tp_periods_choices(form, field, limit=None):
    if form._cw.form.get('vid') == 'holidays_form':
        rset = form._cw.execute("Any C WHERE R use_calendar CU, CU use_calendar C, "
                                "R euser U, U eid %(u)s", {'u': form._cw.user.eid})
        return sorted((v.view('combobox'), six.text_type(v.eid))
                      for v in rset.entities())
    return field.__class__.choices(field, form, limit=limit)


_affk.tag_object_of(('*', 'periods', 'Timeperiod'),
                    {'choices': tp_periods_choices})


def activity_done_by_choices(form, field, limit=None):
    user = form._cw.user
    entity = form.edited_entity
    # managers can edit the done_by relation as they wish
    if user.is_in_group('managers'):
        rql = 'Any R,T '
        if limit is not None:
            rql += 'LIMIT %s ' % limit
        rql += 'WHERE R is Resource, R title T'
        return sorted((entity.view('combobox'), six.text_type(entity.eid))
                      for entity in form._cw.execute(rql).entities())
    # users can't edit an existing done_by relation
    if entity.has_eid():
        return []
    rql = 'Any R,T WHERE R euser U, R title T, U eid %(u)s'
    res = form._cw.execute(rql, {'u': user.eid}).get_entity(0, 0)
    return [(res.view('combobox'), six.text_type(res.eid))]


_affk.tag_subject_of(('Activity', 'done_by', '*'),
                     {'choices': activity_done_by_choices})


def activity_done_for_choices(form, field, limit=None):
    req = form._cw
    user = req.user
    options = []
    open_state = req.vreg['etypes'].etype_class('WorkOrder').open_state
    # managers can edit the done_for relation as they wish
    if user.is_in_group('managers'):
        if limit is None:
            limit = ''
        else:
            limit = ' LIMIT %s' % limit
        rql = ('Any WO,T %s WHERE WO is WorkOrder, WO title T, '
               'WO in_state S, S name "%s"' % (limit, open_state))
        rset = form._cw.execute(rql)
        if rset:
            options += [(req._('workorders in state %s') % req._(open_state), None)]
            options += sorted((entity.view('combobox'), six.text_type(entity.eid))
                              for entity in rset.entities())
        else:
            options += [(req._('no workorders in state %s') % req._(open_state),
                         INTERNAL_FIELD_VALUE)]
    else:
        # for new entities, users will only see their matching resource
        rset = req.execute('DISTINCT Any WO,T WHERE WO is WorkOrder, WO title T, '
                           'WO in_state S, S name %(s)s, '
                           'WO todo_by R, R euser U, U eid %(u)s',
                           {'u': req.user.eid, 's': open_state})
        if rset:
            options += sorted((entity.view('combobox'), six.text_type(entity.eid))
                              for entity in rset.entities())
        else:
            options += [(req._('no workorders'),
                         INTERNAL_FIELD_VALUE)]
    # ensure current value is in the list
    if form.edited_entity.has_eid():
        budget = form.edited_entity.done_for[0]
        cval = six.text_type(budget.eid)
        if not any(val == cval for _, val in options):
            options.insert(0, (budget.view('combobox'), six.text_type(budget.eid)))
    return options


_affk.tag_subject_of(('Activity', 'done_for', '*'),
                     {'choices': activity_done_for_choices})


def commitment_commit_for_choices(form, field, limit=None):
    req = form._cw
    options = []
    rset = req.execute('DISTINCT Any WO,T WHERE '
                       ' WO is WorkOrder, WO title T, '
                       ' WO end_date >= NOW, '
                       ' WO todo_by R, R euser U, U eid %(u)s',
                       {'u': req.user.eid})
    if rset:
        options += sorted((entity.view('combobox'), six.text_type(entity.eid))
                          for entity in rset.entities())
    else:
        options += [(req._('no workorders'),
                     INTERNAL_FIELD_VALUE)]
    # ensure current value is in the list
    if form.edited_entity.has_eid():
        budget = form.edited_entity.commit_for[0]
        cval = six.text_type(budget.eid)
        if all(val != cval for _, val in options):
            options.insert(0, (budget.view('combobox'), six.text_type(budget.eid)))
    return options


_affk.tag_subject_of(('Commitment', 'commit_for', '*'),
                     {'choices': activity_done_for_choices})



import datetime
from collections import defaultdict

from logilab.mtconverter import xml_escape
from logilab.common.date import (date_range, previous_month, next_month,
                                 first_day, last_day)

from cubicweb import _
from cubicweb_web.view import EntityView
from cubicweb.predicates import is_instance, one_line_rset
from cubicweb_web import form
from cubicweb_web.views import json

from cubicweb_calendar.entities import BadCalendar
from cubicweb_calendar.views import get_date_range_from_reqform


class ResourceActivitySubmitFormView(form.FormViewMixIn, EntityView):
    __regid__ = 'activities-submitform'
    __select__ = one_line_rset() & is_instance('Resource')
    domid = 'activityForm'

    def cell_call(self, row, col, **formvalues):
        self._cw.add_css('cubicweb.form.css')
        resource = self.cw_rset.get_entity(row, col)
        if 'diem' in formvalues:
            diem = formvalues['diem']
        else:
            diem = formvalues['diem'] = self.cw_extra_kwargs.get('date') or datetime.date.today()
        formvalues['duration'] = resource.missing_for(diem)
        self.w(u'<h2>%s</h2>' % self._cw._('declare activities').capitalize())
        self._make_activity_form(resource, **formvalues)

    def _make_activity_form(self, resource, **formvalues):
        if resource is None:
            return u'<div class="error">%s</div>' % (
                self._cw._('no resource for user %s') % resource.dc_title())
        display_fields = [('diem', 'subject'), ('duration', 'subject'),
                          ('done_for', 'subject'), ('description', 'subject'),
                          ('done_by', 'subject')]
        vreg = self._cw.vreg
        activity = vreg['etypes'].etype_class('Activity')(self._cw)
        activity.eid = self._cw.varmaker.next()
        form = vreg['forms'].select('edition', self._cw, entity=activity,
                                    display_fields=display_fields,
                                    redirect_path=self._cw.relative_path(False))
        form.form_buttons = form.form_buttons[:1]  # only keep ok button
        renderer = vreg['formrenderers'].select_or_none('htable', self._cw)
        formvalues.setdefault('done_by', resource.eid)
        form.render(formvalues=formvalues, renderer=renderer, w=self.w)


class ResourceMonitorCalendar(EntityView):
    __regid__ = 'activity_calendar'
    __select__ = is_instance('Resource')

    def call(self):
        for row in range(len(self.cw_rset)):
            # only display resource title if there's more than one in the resultset
            if len(self.cw_rset) > 1:
                entity = self.cw_rset.get_entity(row, 0)
                self.w(u'<h2>%s</h2>' % xml_escape(entity.dc_title()))
            self.cell_call(row, 0)

    def cell_call(self, row, col, firstday=None, calid=None):
        self._cw.add_js('cubicweb.ajax.js')
        self._cw.add_css('cubes.calendar.css')
        resource = self.cw_rset.get_entity(row, col)
        calid = calid or 'tid%s' % resource.eid
        if firstday is None:
            firstday, lastday = get_date_range_from_reqform(self._cw.form, autoset_lastday=True)
        else:
            lastday = last_day(firstday)
        firstday = first_day(firstday)
        # make calendar
        caption = '%s %s' % (self._cw._(firstday.strftime('%B').lower()), firstday.year)
        prevurl, nexturl = self._prevnext_links(firstday, resource, calid)
        prevlink = '<a href="%s">&lt;&lt;</a>' % xml_escape(prevurl)
        nextlink = '<a href="%s">&gt;&gt;</a>' % xml_escape(nexturl)

        # build cells
        try:
            celldatas = self._build_activities(resource, firstday, lastday)
        except BadCalendar as exc:  # in case of missing week day information
            self.w(u'<div class="error">%s</div>' % exc)
            return
        # build table/header
        self.w(u'<table id="%s" class="activitiesCal">'
               u'<tr><th class="prev">%s</th>'
               u'<th class="calTitle" colspan="5"><span>%s</span></th>'
               u'<th class="next">%s</th></tr>'
               u'<tr><th>L</th><th>M</th><th>M</th><th>J</th><th>V</th><th>S</th><th>D</th></tr>'
               % (calid, prevlink, caption, nextlink))
        start = firstday - datetime.timedelta(firstday.weekday())
        # date range exclude last day so we should add one day, hence the 7
        stop = lastday + datetime.timedelta(7 - lastday.weekday())
        for curdate in date_range(start, stop):
            if curdate == start or curdate.weekday() == 0:  # sunday
                self.w(u'<tr>')
            self._build_calendar_cell(curdate, celldatas, firstday)
            if curdate.weekday() == 6:
                self.w(u'</tr>')
        self.w(u'</table>')

    def _build_calendar_cell(self, curdate, celldatas, firstday):
        if curdate.month != firstday.month:
            self.w(u'<td class="outofrange"></td>')
        else:
            cssclasses, total_duration, url, workcases = celldatas[curdate]
            total = total_duration
            if workcases:
                total = u"%s (%s)" % (total, workcases)
            cellcontent = u'<a title="total %s" href="%s">%s</a>' % (
                total, url, curdate.day)
            self.w(u'<td class="%s">%s</td>' % (u' '.join(cssclasses),  cellcontent))

    def _prevnext_links(self, curdate, resource, calid):
        prevdate = previous_month(curdate)
        nextdate = next_month(curdate)
        rql = 'Any X WHERE X eid %s' % resource.eid
        prevlink = self._cw.ajax_replace_url(calid, rql=rql, vid='activity_calendar',
                                             replacemode='swap',
                                             firstday=prevdate.strftime('%Y%m01'))
        nextlink = self._cw.ajax_replace_url(calid, rql=rql, vid='activity_calendar',
                                             replacemode='swap',
                                             firstday=nextdate.strftime('%Y%m01'))
        return prevlink, nextlink

    def _build_activities(self, resource, firstday, lastday):
        # build result
        celldatas = {}
        _today = datetime.date.today()
        dtypes = resource.get_day_types(firstday, lastday)
        for date_, dtstate, expected, day_report in resource.iter_activities(firstday, lastday):
            declared = sum([a.duration for a in day_report])
            cssclass = []
            total_duration = u''
            needs_validation = any(a.cw_adapt_to('IWorkflowable').state == 'pending'
                                   for a in day_report)
            try:
                dtype = self._cw.entity_from_eid(dtypes[date_][0])
            except KeyError:
                self.error('no day type on %s for %s (from %s to %s)',
                           resource, date_, firstday, lastday)
                dtype = None
            if dtstate == 'pending':
                cssclass.append(u'daytype_pending')
            if dtype is not None and dtype.day_worked:
                if date_ <= _today:
                    if abs(declared - expected) > 1e-3:
                        cssclass.append(u'problem')
                    elif needs_validation:
                        cssclass.append(u'pending')
                    else:
                        cssclass.append(u'completed')
                total_duration = u'%.2f / %.1f' % (declared, expected)
            else:
                if dtstate != 'pending':
                    cssclass.append(u'closed')
                if declared:
                    cssclass.append(u'problem')
            summary = ','.join(a.workorder.dc_long_title() for a in day_report)
            url = xml_escape(self._cw.build_url('activities/%s/%s' %
                                                (resource.title, date_.strftime('%Y%m%d'))))
            celldatas[date_] = (cssclass, total_duration, url, summary)
        if _today in celldatas:
            # celldatas maps days to tuples (cssclass, duration, url, descr)
            celldatas[_today][0].append(u'today')
        return celldatas


class ResourceMissingActivitiesMixIn(object):

    def missing_activities(self, entity):
        firstday, lastday = get_date_range_from_reqform(self._cw.form,
                                                        autoset_lastday=True)
        for first_day_of_month in date_range(firstday, lastday, incmonth=1):
            last_day_of_month = min(lastday, last_day(first_day_of_month))
            for date_, __, expected, day_report in entity.iter_activities(
                    first_day_of_month, last_day_of_month):
                declared = sum([a.duration for a in day_report])
                if abs(declared - expected) > 1e-3:
                    summary = ','.join(a.workorder.dc_long_title() for a in day_report)
                    yield date_, expected, declared, summary


class JsonResourceMissingActivitiesBoard(json.JsonMixIn, ResourceMissingActivitiesMixIn, EntityView):
    __regid__ = 'missing-activities.json'
    __select__ = is_instance('Resource')

    def call(self):
        result = []
        for entity in self.cw_rset.entities():
            result += self.entity_data(entity)
        self.wdata(result)

    def entity_call(self, entity):
        self.wdata(list(self.entity_data(entity)))

    def entity_data(self, entity):
        for date, expected, declared, summary in self.missing_activities(entity):
            yield {'user': entity.euser[0].login, 'date': date,
                   'expected': expected, 'declared': declared, 'summary': summary}


class ResourceMissingActivitiesBoard(ResourceMissingActivitiesMixIn, EntityView):
    __regid__ = 'missing-activities'
    __select__ = is_instance('Resource')

    def _make_report_table(self, first_day_of_month, lines):
        w = self.w
        w(u'<table class="listing actboard">')
        w(u'<tr><th colspan="4">%s</th></tr>' %
          first_day_of_month.strftime('%B %Y'))
        w(u'<tr><th>%s</th><th>%s</th><th>%s</th><th>%s</th></tr>' % (
                _('date'), _('expected activity report'),
                _('declared activity report'), _('same day activity summary')))
        for day, act_url, expected, declared, summary in lines:
            cssclass = u'missing' if declared < expected else u'toomuch'
            w(u'<tr><td><a href="%s">%s</a></td><td>%s</td><td class="%s">%s</td><td>%s</td></tr>' %
              (xml_escape(act_url),
               self._cw.printable_value('Date', day), expected,
               cssclass, declared, xml_escape(summary)))
        w(u'</table>')

    def entity_call(self, entity):
        self._cw.add_css('cubes.timesheet.css')
        lines_per_month = defaultdict(list)
        for date_, expected, declared, summary in self.missing_activities(entity):
            act_url = self._cw.build_url('activities/%s/%s' %
                                         (entity.title, date_.strftime('%Y%m%d')))
            month = date_.replace(day=1)
            lines_per_month[month].append((date_, act_url, expected, declared, summary))
        if lines_per_month:
            self.w(u'<h2>%s</h2>' % entity.dc_title())
            for month, lines in lines_per_month.items():
                self._make_report_table(month, lines)


class ResourceCommitmentSubmitFormView(form.FormViewMixIn, EntityView):
    __regid__ = 'commitment.submitform'
    __select__ = one_line_rset() & is_instance('Resource')
    domid = 'commitmentForm'

    def cell_call(self, row, col, **formvalues):
        self._cw.add_css('cubicweb.form.css')
        resource = self.cw_rset.get_entity(row, col)
        self.w(u'<h2>%s</h2>' % self._cw._('add commitment').capitalize())
        self._make_commitment_form(resource, **formvalues)

    def _make_commitment_form(self, resource, **formvalues):
        if resource is None:
            return u'<div class="error">%s</div>' % (
                self._cw._('no resource for user %s') % resource.dc_title())
        display_fields = [
            ('commit_by', 'subject'), ('commit_for', 'subject'),
            ('begin_date', 'subject'), ('end_date', 'subject'),
            ('duration', 'subject')]
        vreg = self._cw.vreg
        commitment = vreg['etypes'].etype_class('Commitment')(self._cw)
        commitment.eid = self._cw.varmaker.next()
        form = vreg['forms'].select(
            'edition', self._cw, entity=commitment,
            display_fields=display_fields,
            redirect_path=self._cw.relative_path(False))
        form.form_buttons = form.form_buttons[:1]  # only keep ok button
        renderer = vreg['formrenderers'].select_or_none('htable', self._cw)
        formvalues.setdefault('commit_by', resource.eid)
        form.render(formvalues=formvalues, renderer=renderer, w=self.w)



# pylint: disable-msg=W0622
"""cubicweb-timesheet application packaging information"""

modname = 'timesheet'
distname = 'cubicweb-%s' % modname

numversion = (1, 3, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
description = 'record who did what and when for the CubicWeb framework'
web = 'http://www.cubicweb.org/project/%s' % distname
author = 'Logilab'
author_email = 'contact@logilab.fr'
classifiers = [
           'Environment :: Web Environment',
           'Framework :: CubicWeb',
           'Programming Language :: Python',
           'Programming Language :: JavaScript',
           ]

__depends__ = {'cubicweb': '>= 4.0.0, < 5.0.0',
               'cubicweb-web': '>= 1.0.0, < 2.0.0',
               'cubicweb-calendar':  '>= 1.0.0, < 2.0.0',
               'cubicweb-workorder': '>= 1.0.0, < 2.0.0',
               'cubicweb-rqlcontroller': '>= 1.0.0, < 2.0.0',
               }



"""activity related startup views.

:organization: Logilab
:copyright: 2007-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from logilab.common.date import strptime

from cubicweb import _
from cubicweb.predicates import match_form_params
from cubicweb_web.view import StartupView
from cubicweb_web import httpcache
from cubicweb_web.views import startup


class WorkcaseActivityStatsView(StartupView):
    __regid__ = 'actstats'
    __select__ = StartupView.__select__ & match_form_params('workcase')
    ref_attr = u'workcase'

    def call(self):
        _ = self._cw._
        ref = self._cw.form[self.ref_attr]
        start, stop, restrict = self.build_startstop_restriction()
        rset = self.get_activities(ref, start, stop, restrict)
        self.wtitle(ref, start, stop)
        self.wview('actsummary', rset, 'null')

    def wtitle(self, ref, start, stop):
        tail = ''
        if start:
            tail = '%s %s' % (_('from'), self.format_date(start))
        if stop:
            tail += ' %s %s' % (_('until'), self.format_date(stop))
        self.w(u'<h1>%s - %s</h1>' % (ref, tail))

    def build_startstop_restriction(self):
        restrict = []
        start = self._cw.form.get('start')
        stop = self._cw.form.get('stop')
        # XXX (syt) should use self._cw.parse_datetime(), no? (maybe not if we
        # rely on generated url)
        if start:
            start = strptime(start, '%Y-%m-%d')
            restrict.append('A diem >= %(start)s')
        if stop:
            stop = strptime(stop, '%Y-%m-%d')
            restrict.append('A diem < %(stop)s')
        if restrict:
            return start, stop, ',' + ','.join(restrict)
        return start, stop, ''

    def get_activities(self, ref, start=None, stop=None, restrict=u''):
        rql = ('Any A, DI, R, DESCR, D, WO, L ORDERBY DI WHERE A is Activity, '
               'A done_for WO, O split_into WO, W title %%(title)s, '
               'A duration D, A diem DI, A description DESCR, '
               'A done_by R, R euser U, U login L %s' % restrict)
        return self._cw.execute(rql, {'ref': ref, 'start': start,
                                      'stop': stop, })


class ProjectActivityStatsView(WorkcaseActivityStatsView):
    __regid__ = 'prj-actstats'
    __select__ = StartupView.__select__ & match_form_params('project')
    ref_attr = u'project'

    def get_activities(self, ref, start=None, stop=None, restrict=u''):
        rql = ('Any A, DI, R, DESCR, D, V, L ORDERBY DI WHERE A is Activity, '
               'A done_for W, version_of P, P name %%(ref)s, A duration D, '
               'A diem DI, A description DESCR, A done_by R, R euser U, '
               'U login L %s' % restrict)
        return self._cw.execute(rql, {'ref': ref, 'start': start,
                                      'stop': stop, })


class MyActivities(startup.IndexView):
    http_cache_manager = httpcache.NoHTTPCacheManager

    title = _('my activities')

    def call(self):
        req = self._cw
        _ = req._
        req.add_js(('cubicweb.ajax.js', 'cubicweb.edition.js'))
        req.add_css('cubicweb.form.css')
        iresource = self._cw.user.cw_adapt_to('timesheet.IResource')
        if iresource:
            resource = iresource.resource
            self.w(u'Go to your <a href="%s">resource page (%s)</a>'
                   % (resource.absolute_url(), resource.title))
        # orders
        self.w(u'<h1>%s</h1>' % _('Open workorders'))
        open_state = req.vreg['etypes'].etype_class('WorkOrder').open_state
        rset = req.execute('Any W,O ORDERBY T,TT WHERE O title T, O split_into W, '
                           'W is WorkOrder, W title TT, W in_state S, S name "%s"'
                           % open_state)
        self.wview('progress_table_view', rset, 'null')


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, (MyActivities,))
    vreg.register_and_replace(MyActivities, startup.IndexView)



"""unit tests for get_day_types web service
"""

import json

from cubicweb_web.devtools.testlib import WebCWTC


class JsonTests(WebCWTC):
    def setup_database(self):
        with self.admin_access.web_request() as req:
            # get default french calendar
            defaultcal = req.find(u'Calendar', title=u'Calendrier Francais').one()
            # make 1st of may a non working day
            feast_day = req.create_entity(u'Recurrentday',
                                          day_month=u'05-01',
                                          day_type=req.create_entity(u'Daytype', title=u'1er mai',
                                                                     type=u'dt_nonworking'))
            defaultcal.cw_set(days=feast_day)
            # create a test user and corresponding resource, and make this resource
            # use this calendar
            restype = req.find(u'Resourcetype', title=u'person').one()
            testuser = req.create_entity(u'CWUser', login=u'testuser', upassword=u'testuser',
                                         in_group=req.find(u'CWGroup', name=u'users').one())
            caluse = req.create_entity(u'Calendaruse', use_calendar=defaultcal)
            req.create_entity(u'Resource', title=u'testuser', rate=1,
                              euser=testuser, rtype=restype,
                              use_calendar=caluse)
            req.cnx.commit()

    def test_get_daytypes(self):
        """test daytypes.json web service"""
        with self.admin_access.web_request(
            url="view",
            login="testuser",
            start="2014-04-28",
            stop="2014-05-05",
            vid="daytypes.json",
        ) as req:
            response = self.app_handle_request(req)
            self.assertEqual(json.loads(response),
                             [[u'2014-04-28', u'dt_working'],
                              [u'2014-04-29', u'dt_working'],
                              [u'2014-04-30', u'dt_working'],
                              [u'2014-05-01', u'dt_nonworking'],
                              [u'2014-05-02', u'dt_working'],
                              [u'2014-05-03', u'dt_nonworking'],
                              [u'2014-05-04', u'dt_nonworking'],
                              [u'2014-05-05', u'dt_working']])


if __name__ == '__main__':
    import unittest
    unittest.main()



from yams.buildobjs import (EntityType, RelationDefinition,
                            SubjectRelation, String, Float, Date)

from cubicweb.schema import (
    WorkflowableEntityType, ERQLExpression, RRQLExpression, RQLVocabularyConstraint)


class Activity(WorkflowableEntityType):
    """time someone spent working on something
    """
    __permissions__ = {'read': ('managers', 'users'),
                       'update': ('managers', ERQLExpression('X in_state ST, ST name "pending", X owned_by U')),
                       'delete': ('managers', ERQLExpression('X in_state ST, ST name "pending", X owned_by U')),
                       'add': ('managers', 'users'),
                       }
    duration = Float(required=True, default=1.0)
    diem = Date(default='TODAY', required=True)
    description = String(fulltextindexed=True, maxsize=256)


class Resourcetype(EntityType):
    """see projman"""
    title = String(required=True, maxsize=64)


class Resource(EntityType):
    """see projman"""
    title = String(required=True, unique=True, maxsize=64)
    rate = Float()
    rtype = SubjectRelation('Resourcetype', cardinality='1*')
    use_calendar = SubjectRelation('Calendaruse', cardinality='+?', composite='subject')
    euser = SubjectRelation('CWUser', cardinality='??')


class done_by(RelationDefinition):
    """activity performed by a Resource"""
    __permissions__ = {'read': ('managers', 'users'),
                       'delete': ('managers', RRQLExpression('S in_state ST, ST name "pending", O euser U')),
                       'add': ('managers', RRQLExpression('O euser U'),),
                       }
    subject = 'Activity'
    object = 'Resource'
    cardinality = '1*'
    constraints = [RQLVocabularyConstraint('O euser OU?, OU in_state ST, NOT ST name "deactivated"')]


class done_for(RelationDefinition):
    subject = 'Activity'
    object = 'WorkOrder'
    cardinality = '1*'


class todo_by(RelationDefinition):
    __permissions__ = {'read': ('managers', 'users'),
                       'delete': ('managers', RRQLExpression('S owned_by U')),
                       'add': ('managers', RRQLExpression('S owned_by U')),
                       }
    subject = 'WorkOrder'
    object = 'Resource'
    cardinality = '+*'
    constraints = [RQLVocabularyConstraint('O euser OU?, OU in_state ST, NOT ST name "deactivated"')]


class Commitment(EntityType):
    __permissions__ = {'read': ('managers', 'users'),
                       'update': ('managers', ERQLExpression('X owned_by U')),
                       'delete': ('managers', ERQLExpression('X owned_by U')),
                       'add': ('managers', 'users'),
                       }
    duration = Float(required=True, default=1.0)
    begin_date = Date()
    end_date = Date()


class commit_by(RelationDefinition):
    __permissions__ = {'read': ('managers', 'users'),
                       'delete': ('managers', RRQLExpression('O euser U')),
                       'add': ('managers', RRQLExpression('O euser U'),),
                       }
    subject = 'Commitment'
    object = 'Resource'
    cardinality = '1*'
    constraints = [RQLVocabularyConstraint('O euser OU?, OU in_state ST, NOT ST name "deactivated"')]


class commit_for(RelationDefinition):
    __permissions__ = {'read': ('managers', 'users'),
                       'delete': ('managers', RRQLExpression('O todo_by R, R euser U')),
                       'add': ('managers', RRQLExpression('O todo_by R, R euser U'),),
                       }
    subject = 'Commitment'
    object = 'WorkOrder'
    cardinality = '1*'
    constraints = [RQLVocabularyConstraint('O todo_by R, R euser U')]



from datetime import datetime

for wo in rqliter('WorkOrder X WHERE X modification_date >= %(date)s',
                  {'date': datetime(2010, 10, 15)}).entities():
    wo.update_progress()
commit()



"""ajax / json / jsonp functions"""

from datetime import datetime as pydatetime
import json

import six

from cubicweb_web.view import StartupView
from cubicweb.predicates import match_form_params


def get_day_types(self, resource_title, start, stop):
    """returns a list of couples (date, day_type) for a given resource.

    ``day_type`` corresponds to the ``type`` attribute of the ``Daytype``
    entities manipulated.

    The calendar used if the last created calendar for the corresponding user.
    """
    resource = self._cw.find(u'Resource', title=resource_title).one()
    start = pydatetime.strptime(start, '%Y-%m-%d').date()
    stop = pydatetime.strptime(stop, '%Y-%m-%d').date()
    daytypes = []
    for date, (dtype, state) in sorted(resource.get_day_types(start, stop).items()):
        day_type = self._cw.entity_from_eid(dtype).type
        daytypes.append((date.strftime('%Y-%m-%d'), day_type))
    return daytypes


class DayTypesJsonView(StartupView):
    """Dumps a list of couples (date, day_type) for a given login.

    ``day_type`` corresponds to the ``type`` attribute of the ``Daytype``
    entities manipulated.

    Expected request parameters are :

    - ``login`` (the user's login)
    - ``start`` (format must be %Y-%m-%d)
    - ``stop`` (format must be %Y-%m-%d)

    If ``callback`` request parameter is passed, it's used as json padding
    """
    __regid__ = 'daytypes.json'
    __select__ = StartupView.__select__ & match_form_params('login', 'start', 'stop')
    content_type = 'application/json'
    binary = True
    templatable = False

    def call(self):
        data = get_day_types(self, self._cw.form['login'],
                             self._cw.form['start'],
                             self._cw.form['stop'])
        if 'callback' in self._cw.form:
            self._cw.set_content_type('text/javascript')
            self.w('%s(%s)' % (self._cw.form['callback'].encode('ascii'),
                               six.ensure_binary(json.dumps(data))))
        else:
            self.w(six.ensure_binary(json.dumps(data)))



from datetime import timedelta

from logilab.common.date import date_range, todate

from cubicweb.entities import AnyEntity, fetch_config
from cubicweb_web.view import EntityAdapter
from cubicweb.predicates import is_instance, score_entity

from cubicweb_workorder import entities as workorder
from cubicweb_calendar.entities import intersect


class Activity(AnyEntity):
    __regid__ = 'Activity'
    fetch_attrs, cw_fetch_order = fetch_config(['diem', 'done_for', 'duration'])

    @property
    def user(self):
        return self.done_by[0].user

    def dc_title(self, format="text/plain"):
        duration = self.duration
        if self.user:
            login = self.user.login
        else:
            login = u''
        try:
            return u'%s %s %s %s [%s]' % (
                self._cw.format_date(self.diem), login, duration,
                self.workorder.dc_title(),
                self.cw_adapt_to('IWorkflowable').printable_state)
        except Exception:
            # remote sources unavailable or no related wp
            return u'%s %s %s [%s]' % (
                self._cw.format_date(self.diem), login, duration,
                self.cw_adapt_to('IWorkflowable').printable_state)

    def dc_long_title(self):
        return u'%s %s' % (self.dc_title(), self.description)

    @property
    def workorder(self):
        return self.done_for[0]


class IResource(EntityAdapter):
    __regid__ = 'timesheet.IResource'
    __select__ = is_instance('CWUser') and score_entity(lambda x: x.reverse_euser)

    @property
    def resource(self):
        return self.entity.reverse_euser[0]


AWHERE = ('WHERE A is Activity, A diem DI, A duration DU, A description DE, '
          'A done_by R, A in_state S, A done_for WO, O split_into WO, ')
ADETAILS = 'Any A, DI, R, DU, WO, DE, S ORDERBY DI,R ' + AWHERE
A_BY_R = 'Any R, S, SUM(DU), MIN(DI), MAX(DI) GROUPBY R,S ORDERBY S ' + AWHERE
A_BY_W = 'Any WO, S, SUM(DU), MIN(DI), MAX(DI) GROUPBY WO, S ORDERBY S ' + AWHERE


class Resource(AnyEntity):
    __regid__ = 'Resource'
    rest_attr = 'title'
    fetch_attrs, cw_fetch_order = fetch_config(('title',))

    rql_activities = ADETAILS+'R eid %(eid)s'
    rql_activities_groupby_workorder = A_BY_W + 'R eid %(eid)s'

    def dc_title(self):
        return '%s' % (self.title)

    def dc_long_title(self):
        return '%s (%s)' % (self.title, self.rtype[0].title)

    @property
    def calendars(self):
        return [cuse.use_calendar[0] for cuse in self.use_calendar]

    @property
    def default_calendar(self):
        rset = self._cw.execute('Any C ORDERBY T DESC LIMIT 1 WHERE R use_calendar CU, '
                                'CU use_calendar C, C title T, R eid %(r)s',
                                {'r': self.eid})
        if rset:
            return rset.get_entity(0, 0)
        return None

    @property
    def user(self):
        if self.euser:
            return self.euser[0]
        return None

    def get_day_types(self, start, stop):
        day_types = {}
        cuses = []
        for cuse in self.use_calendar:
            cstart = cuse.start or start
            cstart = todate(cstart)
            cstop = cuse.stop or stop
            cstop = todate(cstop)
            if intersect((start, stop), (cstart, cstop)):
                cuses.append((cstart, cstop, cuse))
        for date in date_range(start, stop + timedelta(days=1)):
            for cstart, cstop, cuse in cuses:
                if cstart <= date <= cstop:
                    _, dtype, dstate = cuse.calendar.get_days_type(date, date)[0]
                    day_types[date] = (dtype, dstate)
                    break
        return day_types

    def activities_summary(self, firstday, lastday):
        activities = {}
        rset = self._cw.execute("Any A,DI,DU,ST,S,W,WO,WR WHERE A is Activity, "
                                "  A in_state ST, ST name S, "
                                "  A diem <= %(l)s, A diem >= %(f)s, "
                                "  A diem DI, A duration DU, "
                                "  A done_by R, R eid %(r)s, "
                                "  A done_for WO, W split_into WO, W title WR",
                                {'r': self.eid, 'f': firstday, 'l': lastday})
        activities = {}
        for activity in rset.entities():
            activities.setdefault(activity.diem, []).append(activity)
        return activities

    def iter_activities(self, firstday, lastday):
        """yields a 4-tuple:
        <date>, <state>, <expected>, <day_report>
        where :
          - <date> is the date of the day
          - <state> is the daytype's status (i.e. 'validated' or 'pending')
          - <expected> is the expected imputations for this date (i.e 0., 0.5 or 1.)
          - <day_report> is the list of declared of activities
        """
        day_types = self.get_day_types(firstday, lastday)
        working_dtype = {}
        for eid, state in set(day_types.values()):
            working_dtype[eid] = self._cw.eid_rset(eid).get_entity(0, 0)
        activities = self.activities_summary(firstday, lastday)
        for date_ in date_range(firstday, lastday + timedelta(days=1)):
            day_report = activities.get(date_, ())
            dtype_eid, dtstate = day_types.get(date_, (None, None))
            if dtype_eid is None:
                expected = 0.
            else:
                expected = working_dtype[dtype_eid].expected_worktime
            yield date_, dtstate, expected, day_report

    def missing_for(self, day):
        """return missing activity declaration for `day`"""
        _, _, expected, day_report = next(self.iter_activities(day, day))
        declared = sum(a.duration for a in day_report)
        missing = expected - declared
        if abs(missing) > 1e-3:
            return missing
        return 0.


class Order(workorder.Order):
    rest_attr = 'title'

    rql_activities_groupby_resource = A_BY_R+'O eid %(eid)s'
    rql_activities = ADETAILS+'O eid %(eid)s'


class WorkOrder(workorder.WorkOrder):

    rql_activities_groupby_resource = A_BY_R+'WO eid %(eid)s'
    rql_activities = ADETAILS+'WO eid %(eid)s'

    open_state = 'in progress'

    # XXX deprecate this in favor of IMileStoneAdapter.contractors()
    def contractors(self):
        return self.todo_by

    def update_progress(self, compute=True):
        if compute:
            done = sum(activity.duration for activity in self.reverse_done_for)
            self.cw_set(progress_done=done,
                        progress_todo=max(0, self.budget - done))
        super(WorkOrder, self).update_progress(compute)

    # number of columns to display
    activities_rql_nb_displayed_cols = 10

    def activities_rql(self, limit=None):
        return ADETAILS+'WO eid %(eid)s'


class TimesheetWorkOrderIMileStoneAdapter(workorder.WorkOrderIMileStoneAdapter):

    def contractors(self):
        return self.entity.todo_by


class TimesheetOrderIMileStoneAdapter(workorder.OrderIMileStoneAdapter):

    def contractors(self):
        return list(set(x for worder in self.entity.split_into for x in worder.todo_by))


def registration_callback(vreg):
    vreg.register_all(globals().values(),
                      __name__,
                      [TimesheetOrderIMileStoneAdapter, TimesheetWorkOrderIMileStoneAdapter])
    vreg.register_and_replace(TimesheetWorkOrderIMileStoneAdapter, workorder.WorkOrderIMileStoneAdapter)
    vreg.register_and_replace(TimesheetOrderIMileStoneAdapter, workorder.OrderIMileStoneAdapter)



from cubicweb.predicates import is_instance
from cubicweb_web.facet import RelationFacet, RangeFacet, DateRangeFacet


class ActivityDurationFacet(RangeFacet):
    __regid__ = 'duration-facet'
    __select__ = RangeFacet.__select__ & is_instance('Activity')
    rtype = 'duration'


class ActivityDiemFacet(DateRangeFacet):
    __regid__ = 'diem-facet'
    __select__ = DateRangeFacet.__select__ & is_instance('Activity')
    rtype = 'diem'


class ActivityWorkorderFacet(RelationFacet):
    __regid__ = 'activity-workorder-facet'
    __select__ = RelationFacet.__select__ & is_instance('Activity')
    rtype = 'done_for'
    target_attr = 'title'
    label_vid = 'textoutofcontext'


class ActivityResourceFacet(RelationFacet):
    __regid__ = 'activity-resource-facet'
    __select__ = RelationFacet.__select__ & is_instance('Activity')
    accepts = ('Activity',)
    rtype = 'done_by'
    target_attr = 'title'



"""template automatic tests"""

from cubicweb_web.devtools.testlib import AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):

    def to_test_etypes(self):
        return set(('Activity', 'Resource', 'Resourcetype'))

    def list_startup_views(self):
        # XXX needs to specify req.form['project'| 'workcase'] to test
        # [prj-]acstats views
        return ('index',)  # 'actstats', 'prj-actstats'


if __name__ == '__main__':
    import unittest
    unittest.main()
